# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reasoning interceptor that strips reasoning from responses and tracks reasoning information."""

import json
import re
import threading
from pathlib import Path
from typing import Optional, final

from pydantic import Field

from nemo_evaluator.adapters.caching.diskcaching import Cache
from nemo_evaluator.adapters.decorators import register_for_adapter
from nemo_evaluator.adapters.types import (
    AdapterGlobalContext,
    AdapterResponse,
    PostEvalHook,
    ResponseInterceptor,
)
from nemo_evaluator.logging import BaseLoggingParams, get_logger


@final
@register_for_adapter(
    name="reasoning",
    description="Processes reasoning content in API responses",
)
class ResponseReasoningInterceptor(ResponseInterceptor, PostEvalHook):
    """Processes reasoning tokens from response. Collects statistics. Strips and/or moves reasoning tokens."""

    class Params(BaseLoggingParams):
        """Configuration parameters for reasoning interceptor."""

        end_reasoning_token: str = Field(
            default="</think>",
            description="Token that marks the end of reasoning section, not used if reasoning_content is provided",
        )
        start_reasoning_token: str | None = Field(
            default="<think>",
            description="Token that marks the start of reasoning section, used for tracking if reasoning has started",
        )
        add_reasoning: bool = Field(
            default=True, description="Whether to add reasoning information"
        )
        migrate_reasoning_content: bool = Field(
            default=False,
            description="If reasoning traces are found in `reasoning_content`, they will be moved to `content` field end surrounded by start_reasoning_token and end_reasoning_token",
        )
        enable_reasoning_tracking: bool = Field(
            default=True, description="Enable reasoning tracking and logging"
        )
        include_if_not_finished: bool = Field(
            default=True,
            description="Include reasoning content if reasoning is not finished (end token not found)",
        )
        stats_file_saving_interval: Optional[int] = Field(
            default=None,
            description="How often (every how many responses) to save stats to a file. If None, stats are only saved via post_eval_hook.",
        )
        enable_caching: bool = Field(
            default=True,
            description="Whether to enable caching of individual request reasoning statistics and aggregated reasoning stats. Useful for resuming interrupted runs.",
        )
        cache_dir: str = Field(
            default="/tmp/reasoning_interceptor",
            description="Custom cache directory for reasoning stats interceptor.",
        )
        logging_aggregated_stats_interval: int = Field(
            default=100,
            description="How often (every how many responses) to log aggregated reasoning statistics. Default is 100.",
        )

    end_reasoning_token: str
    start_reasoning_token: str | None
    add_reasoning: bool
    migrate_reasoning_content: bool
    enable_reasoning_tracking: bool
    include_if_not_finished: bool
    enable_caching: bool
    cache_dir: Optional[str]
    _request_stats_cache: Optional[Cache]
    stats_file_saving_interval: Optional[int]
    logging_aggregated_stats_interval: int

    def __init__(self, params: Params):
        """
        Initialize the reasoning interceptor.

        Args:
            params: Configuration parameters
        """
        self.end_reasoning_token = params.end_reasoning_token
        self.start_reasoning_token = params.start_reasoning_token
        self.add_reasoning = params.add_reasoning
        self.migrate_reasoning_content = params.migrate_reasoning_content
        self.enable_reasoning_tracking = params.enable_reasoning_tracking
        self.include_if_not_finished = params.include_if_not_finished
        self.stats_file_saving_interval = params.stats_file_saving_interval
        # Get logger for this interceptor with interceptor context
        self.logger = get_logger(self.__class__.__name__)

        self.enable_caching = params.enable_caching
        self.cache_dir = params.cache_dir
        self.logging_aggregated_stats_interval = (
            params.logging_aggregated_stats_interval
        )

        self._lock = threading.Lock()

        # Initialize reasoning stats first, before cache loading
        self._reasoning_stats = {
            "total_responses": 0,
            "responses_with_reasoning": 0,
            "reasoning_finished_count": 0,
            "reasoning_started_count": 0,
            "reasoning_unfinished_count": 0,
            "reasoning_finished_ratio": 0,
            "avg_reasoning_words": None,
            "avg_original_content_words": None,
            "avg_updated_content_words": None,
            "max_reasoning_words": None,
            "max_original_content_words": None,
            "max_updated_content_words": None,
            "max_reasoning_tokens": None,
            "avg_reasoning_tokens": None,
            "max_updated_content_tokens": None,
            "avg_updated_content_tokens": None,
            "total_reasoning_words": 0,
            "total_original_content_words": 0,
            "total_updated_content_words": 0,
            "total_reasoning_tokens": 0,
            "total_updated_content_tokens": 0,
        }

        # Initialize cache if enabled
        if self.enable_caching:
            cache_path = Path(self.cache_dir)
            cache_path.mkdir(parents=True, exist_ok=True)
            self._request_stats_cache = Cache(cache_path)
            self._load_aggregated_cached_stats()
        else:
            self._request_stats_cache = None

        self.logger.info(
            "Reasoning interceptor initialized",
            end_reasoning_token=self.end_reasoning_token,
            start_reasoning_token=self.start_reasoning_token,
            add_reasoning=self.add_reasoning,
            migrate_reasoning_content=self.migrate_reasoning_content,
            enable_reasoning_tracking=self.enable_reasoning_tracking,
            include_if_not_finished=self.include_if_not_finished,
            stats_file_saving_interval=self.stats_file_saving_interval,
            enable_caching=self.enable_caching,
            cache_dir=self.cache_dir,
            logging_aggregated_stats_interval=self.logging_aggregated_stats_interval,
        )

    def _load_aggregated_cached_stats(self) -> None:
        """Load aggregated reasoning stats from cache instead of processing individual requests."""
        if not self._request_stats_cache:
            return

        # Try to load aggregated stats
        try:
            aggregated_data = self._request_stats_cache.get(
                "_aggregated_reasoning_stats"
            )
            if aggregated_data:
                if isinstance(aggregated_data, str):
                    aggregated_data = json.loads(aggregated_data)

                # Restore aggregated stats
                self._reasoning_stats.update(aggregated_data)
                self.logger.info(
                    "Loaded aggregated reasoning stats from cache",
                    total_responses=self._reasoning_stats["total_responses"],
                    responses_with_reasoning=self._reasoning_stats[
                        "responses_with_reasoning"
                    ],
                    max_reasoning_words=self._reasoning_stats["max_reasoning_words"],
                )
                return
        except Exception as e:
            self.logger.warning(f"Failed to load aggregated reasoning stats: {e}")

        # if no aggregated stats, process individual cached requests
        cached_keys = [
            k for k in self._request_stats_cache.iterkeys() if not k.startswith("_")
        ]
        if not cached_keys:
            self.logger.info("No cached reasoning stats found to aggregate")
            return

        self.logger.info(
            f"No aggregated stats found, processing {len(cached_keys)} individual cached requests"
        )

        # Aggregate individual cached stats (fallback for older cache format)
        for key in cached_keys:
            try:
                cached_data = self._request_stats_cache.get(key)
                if cached_data:
                    if isinstance(cached_data, str):
                        cached_data = json.loads(cached_data)
                    self._update_reasoning_stats(cached_data)
            except Exception as e:
                self.logger.warning(
                    f"Failed to process cached reasoning stats for key {key}: {e}"
                )

        # Save the aggregated stats for next time
        self._save_aggregated_stats()

        # Log summary of what was loaded
        self.logger.info(
            "Cached reasoning stats loaded and aggregated",
            total_responses=self._reasoning_stats["total_responses"],
            responses_with_reasoning=self._reasoning_stats["responses_with_reasoning"],
            max_reasoning_words=self._reasoning_stats["max_reasoning_words"],
        )

    def _cache_reasoning_stats(
        self, request_id: str, stats: dict[str, any], context: AdapterGlobalContext
    ) -> None:
        """Cache individual reasoning stats by request ID."""
        if not self.enable_caching:
            return
        try:
            stats_json = json.dumps(stats, ensure_ascii=False)
            self._request_stats_cache[request_id] = stats_json

        except Exception as e:
            self.logger.warning(f"Failed to cache reasoning stats: {e}")

    def _save_aggregated_stats(self) -> None:
        """Save aggregated reasoning stats to cache for efficient loading."""
        if not self.enable_caching or self._request_stats_cache is None:
            self.logger.debug(
                "Skipping aggregated stats save: caching disabled or no cache"
            )
            return

        try:
            stats_json = json.dumps(self._reasoning_stats, ensure_ascii=False)
            self._request_stats_cache["_aggregated_reasoning_stats"] = stats_json
            self.logger.debug(
                "Saved aggregated reasoning stats to cache",
                total_responses=self._reasoning_stats["total_responses"],
                responses_with_reasoning=self._reasoning_stats[
                    "responses_with_reasoning"
                ],
                stats_size_bytes=len(stats_json),
            )
        except Exception as e:
            self.logger.warning(f"Failed to save aggregated reasoning stats: {e}")

    def _update_reasoning_stats(self, reasoning_info: dict) -> None:
        """Update reasoning statistics with new data (thread-safe)."""
        with self._lock:
            self._reasoning_stats["total_responses"] += 1

            # Get values
            reasoning_words = reasoning_info.get("reasoning_words", 0)
            original_words = reasoning_info.get("original_content_words", 0)
            updated_words = reasoning_info.get("updated_content_words", 0)
            reasoning_tokens = reasoning_info.get("reasoning_tokens", "unknown")
            updated_content_tokens = reasoning_info.get(
                "updated_content_tokens", "unknown"
            )

            # Increment counters
            if (
                reasoning_words == "unknown"
                and reasoning_info.get("reasoning_started") is True
            ) or (isinstance(reasoning_words, int) and reasoning_words > 0):
                # if reasoning started but not finished, or finished and we have non-zero reasoning words
                self._reasoning_stats["responses_with_reasoning"] += 1
            if reasoning_info.get("reasoning_started") is True:
                self._reasoning_stats["reasoning_started_count"] += 1
                if reasoning_info.get("reasoning_finished"):
                    self._reasoning_stats["reasoning_finished_count"] += 1
                else:
                    self._reasoning_stats["reasoning_unfinished_count"] += 1

            # Update running averages
            for stat_key, value in [
                ("avg_reasoning_words", reasoning_words),
                ("avg_original_content_words", original_words),
                ("avg_updated_content_words", updated_words),
                ("avg_reasoning_tokens", reasoning_tokens),
                ("avg_updated_content_tokens", updated_content_tokens),
            ]:
                if value != "unknown":
                    if self._reasoning_stats[stat_key] is None:
                        self._reasoning_stats[stat_key] = value
                    else:
                        self._reasoning_stats[stat_key] = round(
                            (
                                self._reasoning_stats[stat_key]
                                * (self._reasoning_stats["total_responses"] - 1)
                                + value
                            )
                            / self._reasoning_stats["total_responses"],
                            2,
                        )

            # Update max reasoning words
            for key in [
                "reasoning_words",
                "original_content_words",
                "updated_content_words",
                "reasoning_tokens",
                "updated_content_tokens",
            ]:
                value = reasoning_info.get(key, None)
                if value is not None and value != "unknown":
                    if (
                        self._reasoning_stats[f"max_{key}"] is None
                        or value > self._reasoning_stats[f"max_{key}"]
                    ):
                        self._reasoning_stats[f"max_{key}"] = value

            # Update total statistics
            if reasoning_words != "unknown":
                self._reasoning_stats["total_reasoning_words"] += reasoning_words
            if original_words != "unknown":
                self._reasoning_stats["total_original_content_words"] += original_words
            if updated_words != "unknown":
                self._reasoning_stats["total_updated_content_words"] += updated_words
            if reasoning_tokens != "unknown":
                self._reasoning_stats["total_reasoning_tokens"] += reasoning_tokens
            if updated_content_tokens != "unknown":
                self._reasoning_stats["total_updated_content_tokens"] += (
                    updated_content_tokens
                )

            # Update ratio
            if self._reasoning_stats["responses_with_reasoning"]:
                self._reasoning_stats["reasoning_finished_ratio"] = (
                    self._reasoning_stats["reasoning_finished_count"]
                    / self._reasoning_stats["responses_with_reasoning"]
                )

            # Log aggregated stats at specified interval
            if (
                self._reasoning_stats["total_responses"]
                % self.logging_aggregated_stats_interval
                == 0
            ):
                self.logger.info(**self._reasoning_stats)

    def _process_reasoning_message(
        self, msg: dict, usage: dict = None
    ) -> tuple[dict, dict]:
        """
        Process reasoning in the message and return modified message with reasoning info.

        Args:
            msg: The message object containing content and potentially reasoning_content
            usage: Optional usage data from the response for token tracking

        Returns:
            tuple: (modified_message, reasoning_info) where reasoning_info has keys:
                   reasoning_words, original_content_words, updated_content_words, reasoning_finished, reasoning_started
        """
        modified_msg = msg.copy()
        content = msg.get("content", "")
        updated_content_tokens = "unknown"
        reasoning_tokens = "unknown"

        # Check if reasoning_content exists in the message and is not empty
        if "reasoning_content" in msg and msg["reasoning_content"] is not None:
            reasoning_content = msg["reasoning_content"]
            updated_message_content = content
            reasoning_started = (
                True
                if reasoning_content is not None and reasoning_content.strip() != ""
                else False
            )
            if content.strip() == "":
                reasoning_finished = False
            else:
                reasoning_finished = True
            if usage:
                # First try to get reasoning_tokens directly from usage
                reasoning_tokens = usage.get("reasoning_tokens", "unknown")
                updated_content_tokens = usage.get("content_tokens", "unknown")

                # If not found, check in completion_tokens_details and output_tokens_details
                if reasoning_tokens == "unknown":
                    for key in ["completion_tokens_details", "output_tokens_details"]:
                        if key in usage:
                            details = usage[key]
                            if isinstance(details, dict):
                                reasoning_tokens = details.get(
                                    "reasoning_tokens", "unknown"
                                )
                                if reasoning_tokens != "unknown":
                                    self.logger.debug(
                                        f"Found reasoning_tokens in {key}: {reasoning_tokens}"
                                    )
                                    break

                # Log if reasoning tokens were found
                if reasoning_tokens != "unknown":
                    self.logger.debug(f"Reasoning tokens extracted: {reasoning_tokens}")
                else:
                    self.logger.debug("No reasoning tokens found in usage data")

        else:
            reasoning_finished = False
            if self.start_reasoning_token is not None:
                reasoning_started = self.start_reasoning_token in content
            else:
                reasoning_started = "unknown"
            if self.end_reasoning_token in content:
                reasoning_finished = True
                reasoning_started = True

            # Split content using reasoning token
            if reasoning_finished:
                cleaned_content = self._strip_reasoning(content)
                reasoning_content = content[: content.find(self.end_reasoning_token)]
                updated_message_content = cleaned_content
            else:
                if reasoning_started == "unknown":
                    reasoning_content = "unknown"
                elif reasoning_started:
                    reasoning_content = content
                else:
                    reasoning_content = ""
                if not self.include_if_not_finished:
                    updated_message_content = ""
                else:
                    updated_message_content = content

        # Assign the updated message content
        modified_msg["content"] = updated_message_content

        # Calculate lengths and reasoning status
        if reasoning_content and reasoning_content != "unknown":
            reasoning_words = len(reasoning_content.split())
        elif reasoning_started == "unknown":
            reasoning_words = "unknown"
        else:
            reasoning_words = 0

        reasoning_info = {
            "reasoning_words": reasoning_words,
            "original_content_words": (len(content.split()) if content else 0),
            "updated_content_words": (
                len(modified_msg.get("content", "").split())
                if modified_msg.get("content")
                else 0
            ),
            "reasoning_finished": reasoning_finished,
            "reasoning_started": reasoning_started,
            "reasoning_tokens": reasoning_tokens,
            "updated_content_tokens": updated_content_tokens,
        }

        return modified_msg, reasoning_info

    def _strip_reasoning(self, text: str) -> str:
        """Remove everything between start and end reasoning tokens."""
        if not isinstance(text, str):
            return ""

        # Remove everything between start and end reasoning tokens
        # Also handle cases where only end token is present
        cleaned_content = re.sub(
            r".*?" + re.escape(self.end_reasoning_token),
            "",
            text,
            flags=re.DOTALL,
        ).strip("\n")
        return cleaned_content

    def _migrate_reasoning_content(self, msg: dict):
        """Migrate reasoning content to the content field with reasoning tokens."""
        modified_msg = msg.copy()
        if (
            "reasoning_content" in msg
            and msg["reasoning_content"]
            and msg["reasoning_content"].strip()
        ):
            reasoning_content = msg["reasoning_content"]
            content = msg.get("content", "")
            updated_message_content = (
                self.start_reasoning_token
                + reasoning_content
                + self.end_reasoning_token
                + content
            )
            modified_msg["content"] = updated_message_content
            return modified_msg
        else:
            return msg

    @final
    def intercept_response(
        self, resp: AdapterResponse, context: AdapterGlobalContext
    ) -> AdapterResponse:
        """Remove reasoning tokens from assistant message content in the response and track reasoning info."""
        if not self.add_reasoning:
            if self.migrate_reasoning_content:
                self.logger.debug(
                    "Reasoning processing disabled, but migrating `reasoning_content` back to `content`."
                )
                response_data = resp.r.json()
                for choice in response_data["choices"]:
                    msg = choice.get("message")
                    modified_msg = self._migrate_reasoning_content(msg)
                    msg.update(modified_msg)

                resp.r._content = json.dumps(response_data).encode()
            else:
                self.logger.debug(
                    "Reasoning processing disabled, returning response as-is"
                )
            return resp
        if resp.rctx.cache_hit:
            self.logger.debug("Response was from cache, skipping reasoning processing")
            return resp

        try:
            response_data = resp.r.json()

            if isinstance(response_data, dict) and "choices" in response_data:
                self.logger.debug(
                    "Processing response with choices",
                    choices_count=len(response_data["choices"]),
                )

                # Extract usage data from response
                usage_data = response_data.get("usage", {})

                for choice in response_data["choices"]:
                    msg = choice.get("message")
                    if (
                        msg
                        and msg.get("role") == "assistant"
                        and isinstance(msg.get("content"), str)
                    ):
                        # Get modified message and reasoning information
                        modified_msg, reasoning_info = self._process_reasoning_message(
                            msg, usage_data
                        )

                        # Collect reasoning statistics
                        self._update_reasoning_stats(reasoning_info)

                        # Log reasoning information if tracking is enabled
                        if self.enable_reasoning_tracking:
                            self.logger.info(
                                "Reasoning tracking information", **reasoning_info
                            )

                        # Update the message with the modified content
                        msg.update(modified_msg)

                        # Save stats to file if interval reached
                        if (
                            self.stats_file_saving_interval is not None
                            and self._reasoning_stats["total_responses"]
                            % self.stats_file_saving_interval
                            == 0
                        ):
                            self._save_stats_to_file(context)

                        self.logger.debug(
                            "Message processed",
                            role=msg.get("role"),
                            original_content_length=reasoning_info[
                                "original_content_words"
                            ],
                            updated_content_length=reasoning_info[
                                "updated_content_words"
                            ],
                            reasoning_words=reasoning_info["reasoning_words"],
                        )

                # Cache individual reasoning stats if enabled
                if self.enable_caching:
                    request_id = resp.rctx.request_id
                    if request_id:
                        # Cache the individual reasoning stats
                        self._cache_reasoning_stats(request_id, reasoning_info, context)
                        self.logger.debug(
                            "Cached individual reasoning stats",
                            request_id=request_id,
                            individual_stats=reasoning_info,
                        )
                self._save_aggregated_stats()  # Save aggregated stats periodically

            resp.r._content = json.dumps(response_data).encode()

            self.logger.info(
                "Response reasoning processing completed",
            )

        except Exception as e:
            self.logger.error("Failed to process response reasoning", error=str(e))
            pass

        return resp

    def _save_stats_to_file(self, context: AdapterGlobalContext) -> None:
        """Save current stats to the same file as post-eval hook."""
        # Get stats in a thread-safe manner
        with self._lock:
            stats = self._reasoning_stats.copy()

        # Don't create file if no stats collected
        if stats["total_responses"] == 0:
            self.logger.debug("No reasoning statistics collected, skipping file write")
            return

        # Prepare metrics data under adapter name
        metrics_data = {
            "reasoning": {
                "description": "Reasoning statistics saved during processing",
                **stats,
            }
        }

        with self._lock:
            context.metrics_path.parent.mkdir(parents=True, exist_ok=True)

            # Read existing metrics if file exists
            existing_metrics = {}
            if context.metrics_path.exists():
                try:
                    with open(context.metrics_path, "r") as f:
                        existing_metrics = json.load(f)
                except (json.JSONDecodeError, IOError):
                    pass  # Start fresh if file is corrupted

            # Merge with existing metrics
            merged_metrics = {**existing_metrics, **metrics_data}

            # Write merged metrics to file
            with open(context.metrics_path, "w") as f:
                json.dump(merged_metrics, f, indent=2, ensure_ascii=False)

        self.logger.debug(
            "Saved reasoning stats to file",
            path=str(context.metrics_path),
            response_count=stats["total_responses"],
        )

    def post_eval_hook(self, context: AdapterGlobalContext) -> None:
        """Write collected reasoning statistics to eval_factory_metrics.json."""
        self.logger.info(
            "Writing reasoning statistics to metrics",
            total_responses=self._reasoning_stats["total_responses"],
            output_dir=context.output_dir,
        )
        self._save_stats_to_file(context)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Decorators for registering interceptors and post-evaluation hooks with automatic interface detection."""

import inspect
from typing import Type

from pydantic import BaseModel

from nemo_evaluator.adapters.registry import InterceptorMetadata, InterceptorRegistry
from nemo_evaluator.adapters.types import (
    PostEvalHook,
    RequestInterceptor,
    RequestToResponseInterceptor,
    ResponseInterceptor,
)


def register_for_adapter(
    name: str,
    description: str,
):
    """Decorator to register interceptors or post-evaluation hooks with automatic interface detection.

    Args:
        name: Unique name for the adapter component
        description: Human-readable description of what the component does

    Example:
        from pydantic import BaseModel, Field

        @register_for_adapter(
            name="my_interceptor",
            description="My custom request processing logic"
        )
        class MyInterceptor(RequestInterceptor):
            class Params(BaseModel):
                api_key: str = Field(..., description="API key for authentication")
                timeout: int = Field(default=30, description="Request timeout in seconds")

            def __init__(self, params: Params):
                self.api_key = params.api_key
                self.timeout = params.timeout

            def intercept_request(self, ar: AdapterRequest) -> AdapterRequest:
                # Custom logic here
                return ar

        @register_for_adapter(
            name="my_hook",
            description="My custom post-evaluation hook"
        )
        class MyHook(PostEvalHook):
            class Params(BaseModel):
                output_path: str = Field(..., description="Output path")

            def __init__(self, params: Params):
                self.output_path = params.output_path

            def post_eval_hook(self, context: AdapterGlobalContext) -> None:
                # Custom logic here
                pass
    """

    def decorator(cls):
        # Validate that the class implements at least one interface
        implements_request = issubclass(cls, RequestInterceptor)
        implements_response = issubclass(cls, ResponseInterceptor)
        implements_request_to_response = issubclass(cls, RequestToResponseInterceptor)
        implements_post_eval_hook = issubclass(cls, PostEvalHook)

        if not (
            implements_request
            or implements_response
            or implements_request_to_response
            or implements_post_eval_hook
        ):
            raise ValueError(
                f"Class {cls.__name__} must implement at least one of RequestInterceptor, ResponseInterceptor, RequestToResponseInterceptor, or PostEvalHook"
            )

        # Validate that the class has a Params class
        if not hasattr(cls, "Params"):
            raise ValueError(
                f"Class {cls.__name__} must have a nested Params class that inherits from pydantic.BaseModel"
            )
        params_class = cls.Params
        if not issubclass(params_class, BaseModel):
            raise ValueError(
                f"Class {cls.__name__}.Params must inherit from pydantic.BaseModel, got {type(params_class)}"
            )

        # Validate that __init__ takes only a single Params object
        _validate_init_signature(cls, params_class)

        metadata = InterceptorMetadata(
            name=name,
            description=description,
            interceptor_class=cls,
            init_schema=params_class,
        )

        # Store metadata on the class for discovery
        cls._interceptor_metadata = metadata

        # Register with global registry
        registry = InterceptorRegistry.get_instance()
        registry.register(name, cls, metadata)

        return cls

    return decorator


def _validate_init_signature(cls: Type, params_class: Type[BaseModel]) -> None:
    """Validate that the class __init__ method takes only a single Params object.

    Args:
        cls: The interceptor class to validate
        params_class: The Params class for validation

    Raises:
        ValueError: If the __init__ signature doesn't match the expected format
    """
    sig = inspect.signature(cls.__init__)
    params = list(sig.parameters.keys())

    # Remove 'self' from the list
    if params and params[0] == "self":
        params = params[1:]

    # Check that there's exactly one parameter named 'params' of the correct type
    if len(params) != 1 or params[0] != "params":
        raise ValueError(
            f"Class {cls.__name__}.__init__ must take exactly one parameter named 'params' of type {params_class.__name__}, got {params}"
        )

    # Check the type annotation
    param = sig.parameters["params"]
    if param.annotation != params_class:
        raise ValueError(
            f"Class {cls.__name__}.__init__ parameter 'params' must be annotated as {params_class.__name__}, got {param.annotation}"
        )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .caching_interceptor import CachingInterceptor
from .endpoint_interceptor import EndpointInterceptor
from .logging_interceptor import RequestLoggingInterceptor, ResponseLoggingInterceptor
from .payload_modifier_interceptor import PayloadParamsModifierInterceptor
from .progress_tracking_interceptor import ProgressTrackingInterceptor
from .raise_client_error_interceptor import RaiseClientErrorInterceptor
from .reasoning_interceptor import ResponseReasoningInterceptor
from .response_stats_interceptor import ResponseStatsInterceptor
from .system_message_interceptor import SystemMessageInterceptor

__all__ = [
    "CachingInterceptor",
    "EndpointInterceptor",
    "PayloadParamsModifierInterceptor",
    "ProgressTrackingInterceptor",
    "RaiseClientErrorInterceptor",
    "RequestLoggingInterceptor",
    "ResponseLoggingInterceptor",
    "ResponseReasoningInterceptor",
    "ResponseStatsInterceptor",
    "SystemMessageInterceptor",
]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Custom httpx transport that processes requests through adapter pipeline in client mode."""

import asyncio
import json
import time
from typing import Any, Dict
from urllib.parse import urlparse

import httpx

from nemo_evaluator.adapters.adapter_config import AdapterConfig, InterceptorConfig
from nemo_evaluator.adapters.pipeline import AdapterPipeline
from nemo_evaluator.adapters.types import (
    AdapterGlobalContext,
    AdapterRequest,
    AdapterRequestContext,
    AdapterResponse,
    FatalErrorException,
)
from nemo_evaluator.logging import bind_request_id, get_logger

logger = get_logger(__name__)


def create_async_adapter_http_client(
    endpoint_url: str,
    adapter_config: AdapterConfig | None = None,
    output_dir: str = "./nemo_eval_output",
    base_transport: httpx.AsyncBaseTransport | None = None,
    is_base_url: bool = False,
    model_name: str | None = None,
) -> tuple[httpx.AsyncClient, "AsyncAdapterTransport"]:
    """Create an async httpx client with adapter transport for client-mode evaluation.

    Args:
        endpoint_url: The endpoint URL (required)
        adapter_config: Adapter configuration. If None, creates default with endpoint interceptor.
        output_dir: Directory for output files
        base_transport: Optional base async transport to wrap
        is_base_url: If True, client appends paths. If False, use URL as-is.
        model_name: Optional model name for logging context

    Returns:
        Tuple of (httpx.AsyncClient, AsyncAdapterTransport)
    """
    if adapter_config is None:
        adapter_config = AdapterConfig(
            mode="client",
            interceptors=[InterceptorConfig(name="endpoint", enabled=True, config={})],
            post_eval_hooks=[],
        )

    adapter_transport = AsyncAdapterTransport(
        endpoint_url=endpoint_url,
        adapter_config=adapter_config,
        output_dir=output_dir,
        base_transport=base_transport,
        is_base_url=is_base_url,
        model_name=model_name,
    )

    adapter_http_client = httpx.AsyncClient(transport=adapter_transport)

    enabled_interceptors = [ic.name for ic in adapter_config.interceptors if ic.enabled]
    logger.info(
        "Created async adapter HTTP client",
        interceptors=enabled_interceptors,
        is_base_url=is_base_url,
    )

    return adapter_http_client, adapter_transport


class IterableHeaders:
    """Wrapper to make httpx.Headers iterable as (key, value) tuples."""

    def __init__(self, httpx_headers):
        self._headers = httpx_headers

    def __iter__(self):
        return iter(self._headers.items())

    def __getitem__(self, key):
        return self._headers[key]

    def get(self, key, default=None):
        return self._headers.get(key, default)

    def items(self):
        return self._headers.items()

    def keys(self):
        return self._headers.keys()

    def values(self):
        return self._headers.values()

    def __len__(self):
        return len(self._headers)


class HttpxRequestWrapper:
    """Wrapper to make httpx.Request compatible with flask.Request interface."""

    def __init__(self, httpx_request: httpx.Request, override_url: str | None = None):
        self._request = httpx_request
        self._json_cache = None
        self._override_url = override_url
        self._headers_override = None

        if override_url:
            parsed = urlparse(override_url)
            modified_headers_dict = dict(self._request.headers)
            modified_headers_dict["host"] = parsed.netloc
            self._headers_override = httpx.Headers(modified_headers_dict)

    @property
    def method(self) -> str:
        return self._request.method

    @property
    def url(self) -> str:
        return self._override_url if self._override_url else str(self._request.url)

    @property
    def headers(self):
        headers_to_use = (
            self._headers_override if self._headers_override else self._request.headers
        )
        return IterableHeaders(headers_to_use)

    @property
    def cookies(self):
        return None

    @property
    def json(self) -> Dict[str, Any] | None:
        if self._json_cache is not None:
            return self._json_cache

        content_type = self.headers.get("content-type", "")
        if "application/json" in content_type and self._request.content:
            try:
                self._json_cache = json.loads(self._request.content.decode("utf-8"))
                return self._json_cache
            except (json.JSONDecodeError, UnicodeDecodeError):
                return None
        return None

    def get_json(
        self, force: bool = False, silent: bool = False
    ) -> Dict[str, Any] | None:
        return self.json

    @property
    def path(self) -> str:
        """Return the path component of the URL."""
        return self._request.url.path

    def get_data(self, as_text: bool = False) -> bytes | str:
        """Get request body data, compatible with flask.Request.get_data()."""
        if self._request.content:
            if as_text:
                return self._request.content.decode("utf-8", errors="ignore")
            return self._request.content
        return b"" if not as_text else ""


class RequestsResponseWrapper:
    """Wrapper to make httpx.Response compatible with requests.Response interface."""

    def __init__(self, httpx_response: httpx.Response):
        self._response = httpx_response

    @property
    def status_code(self) -> int:
        return self._response.status_code

    @property
    def reason(self) -> str:
        return self._response.reason_phrase

    @property
    def headers(self):
        return self._response.headers

    @property
    def content(self) -> bytes:
        return self._response.content

    @property
    def text(self) -> str:
        return self._response.text

    def json(self) -> Dict[str, Any]:
        return self._response.json()


class AsyncAdapterTransport(httpx.AsyncBaseTransport):
    """Async httpx transport that processes requests through adapter interceptor pipeline."""

    def __init__(
        self,
        endpoint_url: str,
        adapter_config: AdapterConfig,
        output_dir: str,
        base_transport: httpx.AsyncBaseTransport | None = None,
        is_base_url: bool = False,
        model_name: str | None = None,
    ):
        self.adapter_config = adapter_config
        self.output_dir = output_dir
        self.base_transport = base_transport or httpx.AsyncHTTPTransport()
        self.endpoint_url = endpoint_url
        self.is_base_url = is_base_url
        self.model_name = model_name
        self.pipeline = AdapterPipeline(adapter_config, output_dir, model_name)

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        from nemo_evaluator.logging import bind_model_name

        request_id = bind_request_id()

        # Bind the model name to the logging context if available
        if self.model_name:
            bind_model_name(self.model_name)

        request_logger = get_logger()

        original_url = str(request.url)
        request_url = original_url if self.is_base_url else self.endpoint_url

        request_logger.info(
            "Request started",
            method=request.method,
            url=request_url,
        )

        response = await self._process_request_through_pipeline(
            request, request_url, request_id, request_logger
        )

        return response

    async def _process_request_through_pipeline(
        self,
        request: httpx.Request,
        request_url: str,
        request_id: str,
        request_logger,
    ) -> httpx.Response:
        try:
            global_context = AdapterGlobalContext(
                output_dir=self.output_dir,
                url=request_url,
                model_name=self.model_name,
            )

            wrapped_request = HttpxRequestWrapper(request, override_url=request_url)
            adapter_request = AdapterRequest(
                r=wrapped_request,  # type: ignore
                rctx=AdapterRequestContext(request_id=request_id),
            )

            # Run synchronous pipeline in thread pool
            current_request, adapter_response = await asyncio.to_thread(
                self.pipeline.process_request, adapter_request, global_context
            )

            if adapter_response is None:
                modified_httpx_request = self._adapter_request_to_httpx(
                    current_request, request, request_url
                )

                start_time = time.time()
                httpx_response = await self.base_transport.handle_async_request(
                    modified_httpx_request
                )
                latency_ms = round((time.time() - start_time) * 1000, 2)

                await httpx_response.aread()

                wrapped_response = RequestsResponseWrapper(httpx_response)
                adapter_response = AdapterResponse(
                    r=wrapped_response,  # type: ignore
                    rctx=current_request.rctx,
                    latency_ms=latency_ms,
                )

            current_response = await asyncio.to_thread(
                self.pipeline.process_response, adapter_response, global_context
            )

            return self._adapter_response_to_httpx(current_response, request)

        except FatalErrorException as e:
            request_logger.error(f"Fatal error in adapter pipeline: {e}")
            raise
        except Exception as e:
            request_logger.error(f"Error processing request through adapters: {e}")
            raise

    def _adapter_request_to_httpx(
        self,
        adapter_request: AdapterRequest,
        original_request: httpx.Request,
        normalized_url: str | None = None,
    ) -> httpx.Request:
        json_data = adapter_request.r.get_json()
        content = (
            json.dumps(json_data).encode("utf-8")
            if json_data
            else original_request.content
        )
        url = normalized_url if normalized_url else str(original_request.url)

        return httpx.Request(
            method=adapter_request.r.method,
            url=url,
            headers=dict(adapter_request.r.headers),
            content=content,
        )

    def _adapter_response_to_httpx(
        self, adapter_response: AdapterResponse, original_request: httpx.Request
    ) -> httpx.Response:
        return httpx.Response(
            status_code=adapter_response.r.status_code,
            headers=dict(adapter_response.r.headers),
            content=adapter_response.r.content,
            request=original_request,
        )

    def run_post_eval_hooks(self) -> None:
        self.pipeline.run_post_eval_hooks(url="")

    async def aclose(self) -> None:
        try:
            await asyncio.to_thread(self.run_post_eval_hooks)
        finally:
            await self.base_transport.aclose()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logging interceptors with registry support."""

import threading
from typing import Optional, final

from pydantic import Field

from nemo_evaluator.adapters.decorators import register_for_adapter
from nemo_evaluator.adapters.types import (
    AdapterGlobalContext,
    AdapterRequest,
    AdapterResponse,
    RequestInterceptor,
    ResponseInterceptor,
)
from nemo_evaluator.logging import BaseLoggingParams, get_logger


def _get_safe_headers(headers: dict[str, str]) -> dict[str, str]:
    """Create a copy of headers with authorization redacted."""
    safe_headers = dict(headers)
    for header in safe_headers:
        if header.lower() == "authorization":
            safe_headers[header] = "[REDACTED]"
    return safe_headers


@register_for_adapter(
    name="request_logging",
    description="Logs incoming requests",
)
@final
class RequestLoggingInterceptor(RequestInterceptor):
    """Logs incoming requests."""

    class Params(BaseLoggingParams):
        """Configuration parameters for request logging."""

        log_request_body: bool = Field(
            default=True, description="Whether to log request body"
        )
        log_request_headers: bool = Field(
            default=True, description="Whether to log request headers"
        )
        max_requests: Optional[int] = Field(
            default=2,
            description="Maximum number of requests to log (None for unlimited)",
        )

    log_request_body: bool
    log_request_headers: bool
    max_requests: Optional[int]
    _request_count: int

    def __init__(self, params: Params):
        """
        Initialize the request logging interceptor.

        Args:
            params: Configuration parameters
        """
        self.log_request_body = params.log_request_body
        self.log_request_headers = params.log_request_headers
        self.max_requests = params.max_requests
        self._lock = threading.Lock()
        self._request_count = 0

        # Get logger for this interceptor with interceptor context
        self.logger = get_logger(self.__class__.__name__)

        self.logger.info(
            "Request logging interceptor initialized",
            log_request_body=self.log_request_body,
            log_request_headers=self.log_request_headers,
        )

    @final
    def intercept_request(
        self, ar: AdapterRequest, context: AdapterGlobalContext
    ) -> AdapterRequest:
        """Log the incoming request."""

        # NOTE(agronskiy): to reduce lock contention, stop locking after the
        # count is reached.
        with self._lock:
            if (
                self.max_requests is not None
                and self._request_count >= self.max_requests
            ):
                self.logger.debug(
                    "Request logging limit reached, skipping log",
                    max_requests=self.max_requests,
                    current_count=self._request_count,
                )
                return ar

            self._request_count += 1

        log_data = {
            "method": ar.r.method,
            "url": ar.r.url,
            "path": ar.r.path,
        }

        if self.log_request_headers:
            log_data["headers"] = _get_safe_headers(dict(ar.r.headers))

        if self.log_request_body:
            try:
                log_data["body"] = ar.r.get_json()
            except Exception:
                log_data["body"] = ar.r.get_data().decode("utf-8", errors="ignore")

        # Use standard logging (request_id is automatically included from context)
        self.logger.info("Incoming request", **log_data)

        return ar


@register_for_adapter(
    name="response_logging",
    description="Logs responses",
)
@final
class ResponseLoggingInterceptor(ResponseInterceptor):
    """Logs responses."""

    class Params(BaseLoggingParams):
        """Configuration parameters for response logging."""

        log_response_body: bool = Field(
            default=True, description="Whether to log response body"
        )
        log_response_headers: bool = Field(
            default=True, description="Whether to log response headers"
        )
        max_responses: Optional[int] = Field(
            default=None,
            description="Maximum number of responses to log (None for unlimited)",
        )

    log_response_body: bool
    log_response_headers: bool
    max_responses: Optional[int]
    _response_count: int

    def __init__(self, params: Params):
        """
        Initialize the response logging interceptor.

        Args:
            params: Configuration parameters
        """
        self.log_response_body = params.log_response_body
        self.log_response_headers = params.log_response_headers
        self.max_responses = params.max_responses
        self._response_count = 0
        self._lock = threading.Lock()

        # Get logger for this interceptor with interceptor context
        self.logger = get_logger(self.__class__.__name__)

        self.logger.info(
            "Response logging interceptor initialized",
            log_response_body=self.log_response_body,
            log_response_headers=self.log_response_headers,
            max_responses=self.max_responses,
        )

    @final
    def intercept_response(
        self, resp: AdapterResponse, context: AdapterGlobalContext
    ) -> AdapterResponse:
        """Log the outgoing response."""
        # Check if we should log this response based on max_responses limit
        # NOTE(agronskiy): to reduce lock contention, stop locking after the
        # count is reached.
        with self._lock:
            if (
                self.max_responses is not None
                and self._response_count >= self.max_responses
            ):
                self.logger.debug(
                    "Response logging limit reached, skipping log",
                    max_responses=self.max_responses,
                    current_count=self._response_count,
                )
                return resp

            self._response_count += 1

        log_data = {
            "status_code": resp.r.status_code,
            "reason": resp.r.reason,
        }

        if self.log_response_headers:
            log_data["headers"] = dict(resp.r.headers)

        if self.log_response_body:
            try:
                log_data["body"] = resp.r.json()
            except Exception:
                log_data["body"] = resp.r.text

        # Use standard logging (request_id is automatically included from context)
        self.logger.info("Outgoing response", **log_data)

        return resp



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import importlib
import json
import os
import signal
import sys
from typing import Optional

import psutil
import yaml

from nemo_evaluator.adapters.server import AdapterServerProcess
from nemo_evaluator.api.api_dataclasses import (
    Evaluation,
    EvaluationConfig,
    EvaluationMetadata,
    EvaluationResult,
    EvaluationTarget,
)
from nemo_evaluator.core.input import prepare_output_directory, validate_configuration
from nemo_evaluator.core.resources import (
    aggregate_runtime_metrics,
    monitor_memory_usage,
)
from nemo_evaluator.core.utils import run_command
from nemo_evaluator.logging import get_logger

logger = get_logger(__name__)

__all__ = ["evaluate"]


def parse_output(evaluation: Evaluation) -> EvaluationResult:
    # create a module name that is importable
    output_module = importlib.import_module(f"core_evals.{evaluation.pkg_name}.output")
    return output_module.parse_output(evaluation.config.output_dir)


def evaluate(
    eval_cfg: EvaluationConfig,
    target_cfg: EvaluationTarget,
    metadata: Optional[EvaluationMetadata] = None,
) -> EvaluationResult:
    """
    Run an evaluation using configuration objects.

    Args:
        eval_cfg: Evaluation configuration object containing output directory,
                  parameters, and evaluation type
        target_cfg: Target configuration object containing API endpoint details
                    and adapter configuration

    Returns:
        EvaluationResult: Evaluation results and metadata
    """
    run_config = {
        "config": eval_cfg.model_dump(),
        "target": target_cfg.model_dump(),
    }
    evaluation = validate_configuration(run_config)
    prepare_output_directory(evaluation)

    metadata_block = _persist_metadata_and_build_results_block(
        evaluation.config.output_dir, metadata
    )

    # Check if adapter is in client mode (no server needed)
    use_client_mode = (
        target_cfg.api_endpoint
        and target_cfg.api_endpoint.adapter_config
        and target_cfg.api_endpoint.adapter_config.mode == "client"
    )

    # Track if graceful cleanup has been performed
    cleanup_performed = False

    def run_client_mode_cleanup():
        """Run post-eval hooks for client mode during graceful shutdown."""
        nonlocal cleanup_performed
        if cleanup_performed:
            return

        cleanup_performed = True
        if (
            use_client_mode
            and target_cfg.api_endpoint
            and target_cfg.api_endpoint.adapter_config
        ):
            try:
                from nemo_evaluator.adapters.pipeline import AdapterPipeline

                pipeline = AdapterPipeline(
                    target_cfg.api_endpoint.adapter_config,
                    evaluation.config.output_dir,
                    target_cfg.api_endpoint.model_id,
                )
                pipeline.run_post_eval_hooks(url=target_cfg.api_endpoint.url or "")
                logger.info("Post-eval hooks executed during shutdown")
            except Exception as e:
                logger.error(f"Failed to run post-eval hooks during shutdown: {e}")

    def kill_all(signum=None, frame=None):
        """Kill all processes and exit."""
        logger.critical("FATAL: Terminating all processes...")

        # For SIGTERM (graceful shutdown), run client mode cleanup first
        if signum == signal.SIGTERM:
            run_client_mode_cleanup()

        parent = psutil.Process(os.getpid())  # current process
        children = parent.children(recursive=True)
        for child in children:
            if signum == signal.SIGINT:
                # Send SIGINT to children for immediate termination (skip post-eval hooks)
                child.send_signal(signal.SIGINT)
            else:
                # Send SIGTERM to children for graceful termination (run post-eval hooks)
                child.terminate()

        # Use faster timeout for keyboard interrupt (SIGINT)
        timeout = 1 if signum == signal.SIGINT else 5
        gone, alive = psutil.wait_procs(children, timeout=timeout)
        for child in alive:
            logger.warning(f"Force killing child process {child.pid}")
            child.kill()

        sys.exit(1)

    # Set up signal handlers
    signal.signal(signal.SIGTERM, kill_all)
    signal.signal(signal.SIGINT, kill_all)

    def run_evaluation_core():
        # NOTE: if we use NeMoEvaluatorClient on the benchmark side, there's no need to
        # run adapter server for evaluation and we can use client model here
        if use_client_mode:
            logger.info("Using client mode - skipping adapter server")
            cmd = evaluation.render_command()
            run_command(cmd, verbose=True, propagate_errors=True)
            evaluation_result = parse_output(evaluation)

            # In client mode, explicitly run post-eval hooks after command completes
            # This ensures hooks run reliably from the parent process rather than
            # depending on finalizers in the subprocess during interpreter shutdown
            # Note: cleanup_performed flag prevents double execution if SIGTERM was received
            if not cleanup_performed:
                run_client_mode_cleanup()

            return evaluation_result
        else:
            with AdapterServerProcess(evaluation):
                cmd = evaluation.render_command()
                run_command(cmd, verbose=True, propagate_errors=True)
                evaluation_result = parse_output(evaluation)
                return evaluation_result

    # Get cache directory from caching interceptor configuration
    cache_dir = None
    if (
        target_cfg.api_endpoint
        and target_cfg.api_endpoint.adapter_config
        and target_cfg.api_endpoint.adapter_config.interceptors
    ):
        for interceptor in target_cfg.api_endpoint.adapter_config.interceptors:
            if (
                interceptor.name == "caching"
                and interceptor.enabled
                and interceptor.config
                and interceptor.config.get("cache_dir")
            ):
                cache_dir = interceptor.config["cache_dir"]
                logger.info(f"Using caching interceptor cache_dir: {cache_dir}")
                break

    if not cache_dir:
        logger.info("No cache directory configured, token usage will not be collected")

    evaluation_result, metrics = monitor_memory_usage(
        run_evaluation_core,
        interval_ms=100,
        cache_dir=cache_dir,
        output_dir=evaluation.config.output_dir,
    )

    metrics_path = os.path.join(
        evaluation.config.output_dir, "eval_factory_metrics.json"
    )

    # Read existing metrics if file exists
    existing_metrics = {}
    if os.path.exists(metrics_path):
        try:
            with open(metrics_path, "r") as f:
                existing_metrics = json.load(f)
        except (json.JSONDecodeError, IOError):
            pass  # Start fresh if file is corrupted

    # Aggregate all run data from run_times directory
    aggregated_metrics = aggregate_runtime_metrics(evaluation.config.output_dir)

    if aggregated_metrics:
        runtime = aggregated_metrics.get("runtime_seconds", 0)
        inference_time = aggregated_metrics.get("inference_time_seconds", 0)
        scoring_time = aggregated_metrics.get("scoring_time_seconds", 0)
        logger.info(
            "Aggregated metrics",
            runtime_seconds=runtime,
            inference_time_seconds=inference_time,
            scoring_time_seconds=scoring_time,
            peak_memory_bytes=aggregated_metrics.get("peak_memory_bytes", 0),
            total_runs=aggregated_metrics.get("total_runs", 0),
        )

    # Use aggregated metrics if available, otherwise use current metrics
    final_metrics = aggregated_metrics if aggregated_metrics else metrics

    # Merge with existing metrics, using "evaluation" as the key
    # If evaluation key already exists, merge the metrics instead of overwriting
    if "evaluation" in existing_metrics:
        # Aggregate existing evaluation metrics with new ones
        existing_eval = existing_metrics["evaluation"]
        if isinstance(existing_eval, dict) and isinstance(final_metrics, dict):
            # Merge dictionaries with appropriate aggregation strategy
            merged_eval = existing_eval.copy()
            for key, value in final_metrics.items():
                if (
                    key in merged_eval
                    and isinstance(merged_eval[key], (int, float))
                    and isinstance(value, (int, float))
                ):
                    if key in ["runtime_seconds"]:
                        merged_eval[key] += value
                    elif key in ["peak_memory_bytes", "peak_tree_memory_bytes"]:
                        merged_eval[key] = max(merged_eval[key], value)
                    else:
                        merged_eval[key] += value
                elif key == "end_time":
                    merged_eval[key] = value
                elif key == "start_time":
                    merged_eval[key] = value
                else:
                    merged_eval[key] = value
            merged_metrics = {**existing_metrics, "evaluation": merged_eval}
        else:
            merged_metrics = {**existing_metrics, "evaluation": final_metrics}
    else:
        merged_metrics = {**existing_metrics, "evaluation": final_metrics}

    # Write merged metrics to file
    with open(metrics_path, "w") as f:
        json.dump(merged_metrics, f, indent=2)

    evaluation_result_dict = {
        "git_hash": os.getenv("CORE_EVALS_GIT_HASH"),
        "command": evaluation.render_command(),
        "config": evaluation.config.model_dump(exclude_none=True),
        "target": evaluation.target.model_dump(exclude_none=True),
        "results": evaluation_result.model_dump(exclude_none=True),
        **metadata_block,
    }

    logger.info(yaml.dump(evaluation_result_dict))

    with open(os.path.join(evaluation.config.output_dir, "results.yml"), "w") as f:
        yaml.dump(evaluation_result_dict, f)

    return evaluation_result


def _write_with_versioning_header(
    out_dir: str,
    filename: str,
    payload: dict,
    versioning: dict,
):
    header = (
        "# Generated by nemo-evaluator invocation with the versions of components:\n"
    )
    versioning_yaml = yaml.safe_dump(versioning, sort_keys=False).rstrip("\n")
    header += "\n".join(f"# {line}" for line in versioning_yaml.splitlines()) + "\n"
    path = os.path.join(out_dir, filename)
    with open(path, "w") as f:
        f.write(header)
        yaml.safe_dump(payload, f, sort_keys=False)


def _persist_metadata_and_build_results_block(
    out_dir: str, md: Optional[EvaluationMetadata]
) -> dict:
    """Persist the entire metadata object and return a results.yml block.

    Writes the full metadata payload to `metadata.yaml` with a versioning header.
    Returns "metadata.verioning" block which later will be included into results.
    """
    if not md:
        return {}

    md_modified = copy.deepcopy(md)

    # Build versioning block first so we can include it as a commented header
    # in the persisted metadata file for better provenance/debuggability.
    updated_versioning: dict = dict(md_modified.get("versioning", {}))
    git_hash = os.getenv("CORE_EVALS_GIT_HASH")
    if git_hash:
        updated_versioning["git-hash"] = git_hash
    # TODO(agronskiy): we cannot import top level because due to alphabetic auto-sorting in
    # nemo_evaluator.__init__ this leads to circular imports. The said autosorting cannot
    # yet be ignored per-import since this functionality is not supported yet in
    # ruff.
    from nemo_evaluator import __version__ as nemo_evaluator_version

    updated_versioning["nemo_evaluator"] = nemo_evaluator_version

    # Construct full metadata payload to persist and return, augmenting versioning
    # with inferred fields (git-hash, nemo_evaluator_version).
    md_modified["versioning"] = updated_versioning

    with open(os.path.join(out_dir, "metadata.yaml"), "w") as f:
        yaml.safe_dump(md_modified, f, sort_keys=False)

    # For the results.yaml block we only return versioning to keep uncluttered
    return {
        "metadata": {
            "versioning": updated_versioning,
            "__skipped_fields": "see metadata.yaml for the rest of the fields",
        }
    }



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the HTML template for generating reports.

Note: The template is stored as a string in a Python file as a workaround
because many of our pip packages have issues with including .html files
in their distributions. This allows the template to be properly packaged
and distributed with the rest of the code.
"""

SIMPLE_TEMPLATE = """<!DOCTYPE html>
<html>
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>{% if entries %}Example Request-Response Pairs{% else %}Example Request-Response Pair{% endif %}</title>
    <style>
        /* Container and general styles */
        body {
            font-family: Arial, sans-serif;
            line-height: 1.6;
            margin: 0;
            padding: 20px;
            color: #333;
        }
        .container {
            max-width: 1200px;
            margin: 0 auto;
        }
        h1 {
            color: #2c3e50;
            border-bottom: 2px solid #eee;
            padding-bottom: 10px;
            margin-bottom: 10px;
        }
        .subtitle {
            color: #666;
            font-style: italic;
            margin-bottom: 20px;
            font-size: 1.1em;
        }
        .note {
            color: #546E7A;
            font-weight: bold;
            font-style: normal;
        }
        .cache-info-header {
            background-color: #E8F5E9;
            color: #2E7D32;
            padding: 12px 16px;
            border-radius: 4px;
            margin-bottom: 20px;
            font-size: 0.95em;
            border: 1px solid #C8E6C9;
        }
        .cache-info-header .note {
            color: #546E7A;
            font-weight: bold;
            font-style: normal;
            display: block;
            margin-top: 8px;
        }

        /* Entry styles */
        .entry {
            margin: 20px 0;
            padding: 15px;
            background: #f9f9f9;
            border: 1px solid #ddd;
            border-radius: 5px;
        }
        .cache-info {
            padding: 8px 16px;
            margin-bottom: 8px;
            border-radius: 4px;
            font-family: monospace;
            background-color: #E3F2FD;
            color: #0D47A1;
            font-size: 0.9em;
        }
        .cache-explanation {
            font-size: 0.8em;
            color: #666;
            margin-top: 4px;
            font-style: italic;
        }
        .request, .response, .curl-command {
            padding: 8px 16px;
            margin-bottom: 8px;
            border-radius: 4px;
            font-family: monospace;
        }
        .request {
            background-color: #B2DFDB;
            color: #00695C;
        }
        .response {
            background-color: #B39DDB;
            color: #4527A0;
        }
        .curl-command {
            background-color: #EEEEEE;
            color: #212121;
            white-space: pre-wrap;
            display: none;
        }
        pre {
            white-space: pre-wrap;
            margin: 0;
            background: #f5f5f5;
            padding: 10px;
            border-radius: 3px;
            overflow-x: auto;
        }
        .buttons {
            display: flex;
            gap: 10px;
            margin-bottom: 10px;
        }
        .button {
            padding: 5px 10px;
            border: none;
            border-radius: 3px;
            cursor: pointer;
            font-size: 12px;
            background-color: #4CAF50;
            color: white;
        }
        .button:hover {
            background-color: #45a049;
        }
    </style>
</head>
<body>
    <div class="container">
        {% if entries %}
            <h1>Example Request-Response Pairs</h1>
            <p class="subtitle">Final request (sent directly to the endpoint) and response pairs with curl commands for reproducibility. <span class="note">*Note: Responses may vary due to generation parameters, randomness, and other factors.</span></p>
            <div class="cache-info-header">
                <strong>About Cache Keys:</strong> Each request-response pair is identified by a unique cache key generated from the request payload hash. <span class="note">Note: Cache keys may differ across different models because the model field in the request payload will be different.</span>
            </div>
            {% for entry in entries %}
                <div class="entry">
                    <div class="cache-info">
                        <strong>Cache Key:</strong> {{ entry.cache_key }}
                    </div>
                    <div class="request">
                        <strong>Request:</strong>
                        <pre>{{ entry.display_request|tojson_utf8|safe }}</pre>
                    </div>

                    <div class="response">
                        <strong>Response:</strong>
                        <pre>{{ entry.response|tojson_utf8|safe }}</pre>
                    </div>

                    <div class="buttons">
                        <button class="button" onclick="toggleRaw('curl-{{ entry.cache_key }}')">Show Curl Command</button>
                    </div>

                    <div id="curl-{{ entry.cache_key }}" class="curl-command">
                        # Save payload to file first:
                        echo '{{ entry.request_data|tojson }}' > request.json

                        curl "{{ entry.endpoint }}" \
                        -H "Content-Type: application/json" \
                        -H "Authorization: Bearer $API_KEY" \
                        -H "Accept: application/json" \
                        -d @request.json
                    </div>
                </div>
            {% endfor %}
        {% else %}
            <h1>Example Request-Response Pair</h1>
            <p class="subtitle">Final request (sent directly to the endpoint) and response with curl command for reproducibility. <span class="note">*Note: Responses may vary due to generation parameters, randomness, and other factors.</span></p>
            <div class="cache-info-header">
                <strong>About Cache Keys:</strong> Each request-response pair is identified by a unique cache key generated from the request payload hash. <span class="note">Note: Cache keys may differ across different models because the model field in the request payload will be different.</span>
            </div>
            <div class="entry">
                <div class="cache-info">
                    <strong>Cache Key:</strong> {{ cache_key }}
                </div>
                <div class="request">
                    <strong>Request:</strong>
                    <pre>{{ display_request|tojson_utf8|safe }}</pre>
                </div>

                <div class="response">
                    <strong>Response:</strong>
                    <pre>{{ response|tojson_utf8|safe }}</pre>
                </div>

                <div class="buttons">
                    <button class="button" onclick="toggleRaw('curl-{{ cache_key }}')">Show Curl Command</button>
                </div>

                <div id="curl-{{ cache_key }}" class="curl-command">
                        # Save payload to file first:
                        echo '{{ request_data|tojson }}' > request.json

                        curl "{{ endpoint }}" \
                        -H "Content-Type: application/json" \
                        -H "Authorization: Bearer $API_KEY" \
                        -H "Accept: application/json" \
                        -d @request.json
                </div>
            </div>
        {% endif %}
    </div>

    <script>
        function toggleRaw(elementId) {
            const element = document.getElementById(elementId);
            if (element.style.display === 'none' || element.style.display === '') {
                element.style.display = 'block';
            } else {
                element.style.display = 'none';
            }
        }
    </script>
</body>
</html>"""



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Caching interceptor with registry support."""

import hashlib
import json
import threading
from typing import Any, final

import requests
import requests.structures
from pydantic import Field

from nemo_evaluator.adapters.caching.diskcaching import Cache
from nemo_evaluator.adapters.decorators import register_for_adapter
from nemo_evaluator.adapters.types import (
    AdapterGlobalContext,
    AdapterRequest,
    AdapterResponse,
    RequestToResponseInterceptor,
    ResponseInterceptor,
)
from nemo_evaluator.logging import BaseLoggingParams, get_logger


@register_for_adapter(
    name="caching",
    description="Caches requests and responses with disk storage",
)
@final
class CachingInterceptor(RequestToResponseInterceptor, ResponseInterceptor):
    """Caching interceptor is special in the sense that it intercepts both requests and responses."""

    class Params(BaseLoggingParams):
        """Configuration parameters for caching."""

        cache_dir: str = Field(
            default="/tmp", description="Directory to store cache files"
        )
        reuse_cached_responses: bool = Field(
            default=False,
            description="Whether to reuse cached responses. If True, this overrides save_responses (sets it to True) and max_saved_responses (sets it to None)",
        )
        save_requests: bool = Field(
            default=False, description="Whether to save requests to cache"
        )
        save_responses: bool = Field(
            default=True,
            description="Whether to save responses to cache. Note: This is automatically set to True if reuse_cached_responses is True",
        )
        max_saved_requests: int | None = Field(
            default=None, description="Maximum number of requests to save"
        )
        max_saved_responses: int | None = Field(
            default=None,
            description="Maximum number of responses to cache. Note: This is automatically set to None if reuse_cached_responses is True",
        )

    responses_cache: Cache
    requests_cache: Cache
    headers_cache: Cache

    def __init__(self, params: Params):
        """
        Initialize the caching interceptor.

        Args:
            params: Configuration parameters
        """

        # Initialize caches immediately
        self.responses_cache = Cache(directory=f"{params.cache_dir}/responses")
        self.requests_cache = Cache(directory=f"{params.cache_dir}/requests")
        self.headers_cache = Cache(directory=f"{params.cache_dir}/headers")
        self.reuse_cached_responses = params.reuse_cached_responses
        self.save_requests = params.save_requests

        # If reuse_cached_responses is True, override save_responses and max_saved_responses
        if params.reuse_cached_responses:
            self.save_responses = True
            self.max_saved_responses = None
        else:
            self.save_responses = params.save_responses
            self.max_saved_responses = params.max_saved_responses

        self.max_saved_requests = params.max_saved_requests

        # Counters for cache management
        self._cached_requests_count = 0
        self._cached_responses_count = 0

        # Thread safety
        self._count_lock = threading.Lock()

        # Get logger for this interceptor with interceptor context
        self.logger = get_logger(self.__class__.__name__)

        self.logger.info(
            "Caching interceptor initialized",
            cache_dir=params.cache_dir,
            reuse_cached_responses=self.reuse_cached_responses,
            save_requests=self.save_requests,
            save_responses=self.save_responses,
            max_saved_requests=self.max_saved_requests,
            max_saved_responses=self.max_saved_responses,
        )

    @staticmethod
    def _generate_cache_key(data: Any) -> str:
        """
        Generate a hash for the request data to be used as the cache key.

        Args:
            data: Data to be hashed

        Returns:
            str: Hash of the data
        """
        data_str = json.dumps(data, sort_keys=True)
        return hashlib.sha256(data_str.encode("utf-8")).hexdigest()

    def _get_from_cache(self, cache_key: str) -> tuple[Any, Any] | None:
        """
        Attempt to retrieve content and headers from cache.

        Args:
            cache_key (str): Cache key to lookup

        Returns:
            Optional[tuple[Any, Any]]: Tuple of (content, headers) if found, None if not
        """
        try:
            cached_content = self.responses_cache[cache_key]
            cached_headers = self.headers_cache[cache_key]
            self.logger.debug("Cache hit", cache_key=cache_key[:8] + "...")
            return cached_content, cached_headers
        except KeyError:
            self.logger.debug("Cache miss", cache_key=cache_key[:8] + "...")
            return None

    def _save_to_cache(self, cache_key: str, content: Any, headers: Any) -> None:
        """
        Save content and headers to cache.

        Args:
            cache_key (str): Cache key to store under
            content: Content to cache
            headers: Headers to cache
        """
        # Check if we've reached the max responses limit
        if self.max_saved_responses is not None:
            with self._count_lock:
                if self._cached_responses_count >= self.max_saved_responses:
                    self.logger.warning(
                        "Maximum cached responses limit reached",
                        max_saved_responses=self.max_saved_responses,
                    )
                    return
                self._cached_responses_count += 1

        # Save content to cache
        self.responses_cache[cache_key] = content

        # NOTE: headers are `CaseInsensitiveDict()` which is not serializable
        # by `Cache` class. If this is the case, transform to a normal dict.
        if isinstance(headers, requests.structures.CaseInsensitiveDict):
            cached_headers = dict(headers)
        else:
            cached_headers = headers
        self.headers_cache[cache_key] = cached_headers

        self.logger.debug(
            "Saved response to cache",
            cache_key=cache_key[:8] + "...",
            content_size=len(content) if hasattr(content, "__len__") else "unknown",
        )

    @final
    def intercept_request(
        self, req: AdapterRequest, context: AdapterGlobalContext
    ) -> AdapterRequest | AdapterResponse:
        """Shall return request if no cache hit, and response if it is.
        Args:
            req (AdapterRequest): The adapter request to intercept
            context (AdapterGlobalContext): Global context containing server-level configuration
        """
        request_data = req.r.get_json()

        # Check cache. Create cache key that will be used everywhere (also if no cache hit)
        req.rctx.cache_key = self._generate_cache_key(request_data)
        self.logger.debug("Request", request_data=request_data)
        self.logger.debug(
            "Processing request for caching",
            cache_key=req.rctx.cache_key[:8] + "...",
            request_data_keys=(
                list(request_data.keys())
                if isinstance(request_data, dict)
                else "unknown"
            ),
        )

        # Cache request if needed and within limit
        if self.save_requests:
            with self._count_lock:
                if (
                    self.max_saved_requests is None
                    or self._cached_requests_count < self.max_saved_requests
                ):
                    self.requests_cache[req.rctx.cache_key] = request_data
                    self._cached_requests_count += 1
                    self.logger.debug(
                        "Saved request to cache",
                        cache_key=req.rctx.cache_key[:8] + "...",
                    )
                else:
                    self.logger.warning(
                        "Maximum cached requests limit reached",
                        max_saved_requests=self.max_saved_requests,
                    )

        # Only check cache if response reuse is enabled
        if self.reuse_cached_responses:
            cached_result = self._get_from_cache(req.rctx.cache_key)
            if cached_result:
                content, headers = cached_result

                requests_response = requests.Response()
                requests_response._content = content
                requests_response.status_code = 200
                requests_response.reason = "OK"
                requests_response.headers = requests.utils.CaseInsensitiveDict(headers)
                requests_response.request = request_data

                # Make downstream know
                req.rctx.cache_hit = True

                self.logger.info(
                    "Returning cached response",
                    cache_key=req.rctx.cache_key[:8] + "...",
                    status_code=200,
                )

                return AdapterResponse(r=requests_response, rctx=req.rctx)

        self.logger.debug(
            "No cache hit, proceeding with request",
            cache_key=req.rctx.cache_key[:8] + "...",
        )
        return req

    @final
    def intercept_response(
        self, resp: AdapterResponse, context: AdapterGlobalContext
    ) -> AdapterResponse:
        """Cache the response if caching is enabled and response is successful."""

        # first, if caching was used, we do nothing
        if resp.rctx.cache_hit:
            self.logger.debug(
                "Response was from cache, skipping caching",
                cache_key=(
                    resp.rctx.cache_key[:8] + "..."
                    if hasattr(resp.rctx, "cache_key")
                    else "unknown"
                ),
            )
            return resp

        if resp.r.status_code == 200 and self.save_responses:
            # Save both content and headers to cache
            try:
                assert resp.rctx.cache_key, "cache key is unset, this is a bug"
                self._save_to_cache(
                    cache_key=resp.rctx.cache_key,
                    content=resp.r.content,
                    headers=resp.r.headers,
                )
                self.logger.info(
                    "Cached successful response",
                    cache_key=resp.rctx.cache_key[:8] + "...",
                )
            except Exception as e:
                self.logger.error(
                    "Could not cache response",
                    error=str(e),
                    cache_key=(
                        resp.rctx.cache_key[:8] + "..."
                        if hasattr(resp.rctx, "cache_key")
                        else "unknown"
                    ),
                )
        else:
            self.logger.debug(
                "Response not cached",
                status_code=resp.r.status_code,
                save_responses=self.save_responses,
            )

        # And just propagate
        return resp



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import os
import sqlite3
import threading
import time
from pathlib import Path
from typing import Any

import psutil

from nemo_evaluator.api.api_dataclasses import EvaluationResult
from nemo_evaluator.logging.utils import logger

__all__ = []


def get_token_usage_from_cache_db(cache_db_path: str | Path) -> dict:
    """Extract token usage metrics from the cache database.

    Args:
        cache_db_path: Path to the SQLite cache database file

    Returns:
        Dictionary containing token usage metrics:
        {
            "total_tokens": int,
            "prompt_tokens": int,
            "completion_tokens": int,
            "total_cached_requests": int
        }
        Returns empty dict if no data found or error occurs.
    """
    try:
        with sqlite3.connect(cache_db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(
                """
                SELECT
                    SUM(json_extract(value, '$.usage.total_tokens')) as total_tokens,
                    SUM(json_extract(value, '$.usage.prompt_tokens')) as prompt_tokens,
                    SUM(json_extract(value, '$.usage.completion_tokens')) as completion_tokens,
                    COUNT(*) as total_requests
                FROM cache
                WHERE json_extract(value, '$.usage') IS NOT NULL
            """
            )
            row = cursor.fetchone()

            if row and row[0] is not None:  # Check if we got any results
                return {
                    "total_tokens": row[0],
                    "prompt_tokens": row[1],
                    "completion_tokens": row[2],
                    "total_cached_requests": row[3],
                }
    except Exception as e:
        logger.warning("Failed to read token usage from cache", error=str(e))

    return {}


def get_token_usage_from_cache(cache_dir: str) -> dict:
    """Extract token usage metrics from the cache database."""
    cache_db_path = Path(cache_dir) / "responses" / "cache.db"
    if not cache_db_path.exists():
        return {}

    return get_token_usage_from_cache_db(cache_db_path)


def aggregate_runtime_metrics(output_dir: str) -> dict[str, Any]:
    """Aggregate all run data from run_times directory."""
    run_times_dir = Path(output_dir) / "run_times"
    aggregated_metrics = {}

    if not run_times_dir.exists():
        return aggregated_metrics

    total_runtime = 0
    earliest_start = None
    latest_end = None
    max_peak_memory = 0
    max_peak_tree_memory = 0
    run_count = 0

    for run_file in run_times_dir.glob("runtime_*.json"):
        try:
            with open(run_file, "r") as f:
                run_data = json.load(f)
                total_runtime += run_data.get("runtime_seconds", 0)
                run_count += 1

                # Track earliest start and latest end
                run_start = run_data.get("start_time", "")
                run_end = run_data.get("end_time", "")
                if earliest_start is None or run_start < earliest_start:
                    earliest_start = run_start
                if latest_end is None or run_end > latest_end:
                    latest_end = run_end

                # Track peak memory across all runs
                max_peak_memory = max(
                    max_peak_memory, run_data.get("peak_memory_bytes", 0)
                )
                max_peak_tree_memory = max(
                    max_peak_tree_memory, run_data.get("peak_tree_memory_bytes", 0)
                )
        except Exception:
            pass

    if run_count > 0:
        aggregated_metrics = {
            "runtime_seconds": total_runtime,
            "start_time": earliest_start,
            "end_time": latest_end,
            "peak_memory_bytes": max_peak_memory,
            "peak_tree_memory_bytes": max_peak_tree_memory,
            "total_runs": run_count,
        }

        # Try to get inference time from response stats and calculate scoring time
        try:
            metrics_file = Path(output_dir) / "eval_factory_metrics.json"
            if metrics_file.exists():
                with open(metrics_file, "r") as f:
                    metrics_data = json.load(f)
                    response_stats = metrics_data.get("response_stats", {})
                    inference_time = response_stats.get("inference_time", 0.0)

                    # Calculate scoring time as runtime - inference time
                    scoring_time = max(0.0, total_runtime - inference_time)
                    aggregated_metrics["inference_time_seconds"] = inference_time
                    aggregated_metrics["scoring_time_seconds"] = scoring_time
        except Exception as e:
            # If we can't read response stats, just continue without scoring time
            logger.warning(
                "Could not extract inference time from response stats", error=str(e)
            )

    return aggregated_metrics


def _update_persistent_metrics(
    output_dir: str,
    start_time: float,
    peak_memory: int,
    peak_tree_memory: int,
    run_id: str,
) -> None:
    """Save individual run data and update peak memory only."""
    try:
        # Create run_times directory
        run_times_dir = Path(output_dir) / "run_times"
        run_times_dir.mkdir(exist_ok=True)

        # Save individual run runtime
        current_time = time.time()
        current_runtime = current_time - start_time
        run_file = run_times_dir / f"runtime_{run_id}.json"

        with open(run_file, "w") as f:
            json.dump(
                {
                    "run_id": run_id,
                    "start_time": time.strftime(
                        "%Y-%m-%dT%H:%M:%S.%fZ", time.gmtime(start_time)
                    ),
                    "end_time": time.strftime(
                        "%Y-%m-%dT%H:%M:%S.%fZ", time.gmtime(current_time)
                    ),
                    "runtime_seconds": current_runtime,
                    "peak_memory_bytes": peak_memory,
                    "peak_tree_memory_bytes": peak_tree_memory,
                },
                f,
            )

    except Exception as e:
        logger.warning(
            "Failed to update persistent metrics", error=str(e), run_id=run_id
        )


def monitor_memory_usage(
    func,
    *args,
    interval_ms,
    cache_dir: str | None = None,
    output_dir: str | None = None,
    **kwargs,
) -> tuple[EvaluationResult, dict[str, Any]]:
    """
    Run func(*args, **kwargs) while polling RSS via psutil.
    Returns (func_return_value, peak_rss_bytes, peak_tree_rss_bytes) where:
    - peak_rss_bytes: peak memory usage of the main process
    - peak_tree_rss_bytes: peak memory usage of the entire process tree (main + children)
    """
    proc = psutil.Process(os.getpid())

    # Generate meaningful run ID (counter or date)
    if output_dir:
        run_times_dir = Path(output_dir) / "run_times"
        run_times_dir.mkdir(exist_ok=True)
        # Count existing runs to get next ID
        existing_runs = list(run_times_dir.glob("runtime_*.json"))
        run_id = str(len(existing_runs))
    else:
        run_id = "0"

    # Initialize values
    peak = 0
    peak_tree = 0

    stop = False
    ret = None

    def get_tree_memory(process):
        """Get total memory usage of process and all its children."""
        try:
            memory = process.memory_info().rss
            for child in process.children(recursive=True):
                try:
                    memory += child.memory_info().rss
                except (psutil.NoSuchProcess, psutil.AccessDenied):
                    pass
            return memory
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            return 0

    def sampler():
        nonlocal peak, peak_tree
        last_save_time = 0
        save_interval = 5.0  # Save every 5 seconds

        while not stop:
            # Get memory for current process
            rss = proc.memory_info().rss
            peak = max(peak, rss)

            # Get memory for entire process tree
            tree_rss = get_tree_memory(proc)
            peak_tree = max(peak_tree, tree_rss)

            # Update persistent metrics file if output_dir is provided and enough time has passed
            if output_dir:
                current_time = time.time()
                if current_time - last_save_time >= save_interval:
                    _update_persistent_metrics(
                        output_dir, start_time, peak, peak_tree, run_id
                    )
                    last_save_time = current_time

            time.sleep(interval_ms / 1000.0)

    th = threading.Thread(target=sampler, daemon=True)
    start_time = time.time()
    th.start()

    try:
        # Filter out cache_dir from kwargs since the target function doesn't expect it
        func_kwargs = {k: v for k, v in kwargs.items() if k != "cache_dir"}
        ret = func(*args, **func_kwargs)
    finally:
        stop = True  # thread safe
        th.join()
    end_time = time.time()

    runtime_seconds = end_time - start_time

    # Get token usage from cache if cache_dir is provided
    token_usage = None
    if cache_dir:
        try:
            token_usage = get_token_usage_from_cache(cache_dir)
        except Exception as e:
            logger.warning("Failed to get token usage from cache", error=str(e))

    metrics = {
        "runtime_seconds": runtime_seconds,
        "start_time": time.strftime("%Y-%m-%dT%H:%M:%S.%fZ", time.gmtime(start_time)),
        "end_time": time.strftime("%Y-%m-%dT%H:%M:%S.%fZ", time.gmtime(end_time)),
        "token_usage": token_usage,
        "peak_memory_bytes": peak,
        "peak_tree_memory_bytes": peak_tree,
    }

    return ret, metrics



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import os
import subprocess
import tempfile
import time
from typing import Any, Literal, TypeVar

import requests
import yaml
from jinja2 import Environment, StrictUndefined, nodes

from nemo_evaluator.logging import get_logger

__all__ = []

logger = get_logger(__name__)


class MisconfigurationError(Exception):
    pass


KeyType = TypeVar("KeyType")


def get_jinja2_environment() -> Environment:
    """Get a configured Jinja2 environment for template operations.

    This ensures consistency between template parsing and rendering.
    Uses StrictUndefined to match the behavior in api_dataclasses.py.

    Returns:
        Environment: Configured Jinja2 environment
    """
    return Environment(undefined=StrictUndefined)


def deep_update(
    mapping: dict[KeyType, Any],
    *updating_mappings: dict[KeyType, Any],
    skip_nones: bool = False,
) -> dict[KeyType, Any]:
    """Deep update a mapping with other mappings.

    If `skip_nones` is True, then the values that are None in the updating mappings are
    not updated.
    """
    updated_mapping = mapping.copy()
    for updating_mapping in updating_mappings:
        for k, v in updating_mapping.items():
            if (
                k in updated_mapping
                and isinstance(updated_mapping[k], dict)
                and isinstance(v, dict)
            ):
                updated_mapping[k] = deep_update(
                    updated_mapping[k], v, skip_nones=skip_nones
                )
            else:
                if skip_nones and v is None:
                    continue
                updated_mapping[k] = v
    return updated_mapping


def extract_params_from_command(command: str) -> tuple[set[str], set[str]]:
    """Extract all config.params.* parameter names used in a command template.

    Args:
        command: Jinja2 command template string

    Returns:
        Tuple of (standard_params, extra_params) where:
        - standard_params: Set of param names like {'temperature', 'max_new_tokens'}
        - extra_params: Set of params.extra names like {'dummy_score', 'another_param'}
    """
    # Use Jinja2's AST parser to extract variable attribute access patterns
    # Uses the same environment configuration as template rendering
    env = get_jinja2_environment()
    ast = env.parse(command)

    standard_params = set()
    extra_params = set()

    def extract_getattr_path(node):
        """Recursively extract the full dotted path from a Getattr node."""
        if isinstance(node, nodes.Name):
            return node.name
        elif isinstance(node, nodes.Getattr):
            base = extract_getattr_path(node.node)
            if base:
                return f"{base}.{node.attr}"
            return node.attr
        return None

    def visit_node(node):
        """Visit all nodes in the AST to find variable references."""
        if isinstance(node, nodes.Getattr):
            full_path = extract_getattr_path(node)
            if full_path:
                # Check for config.params.extra.PARAM_NAME pattern
                if full_path.startswith("config.params.extra."):
                    param_name = full_path.replace("config.params.extra.", "")
                    if param_name:  # Only add if there's something after "extra."
                        # Extract only the first-level key after "extra."
                        param_name = param_name.split(".")[0]
                        extra_params.add(param_name)
                # Check for config.params.PARAM_NAME pattern (but not just config.params.extra)
                elif (
                    full_path.startswith("config.params.")
                    and full_path != "config.params.extra"
                ):
                    param_name = full_path.replace("config.params.", "")
                    if param_name != "extra":  # Don't add "extra" itself
                        standard_params.add(param_name)

        # Recursively visit child nodes
        for child in node.iter_child_nodes():
            visit_node(child)

    visit_node(ast)
    return standard_params, extra_params


def validate_params_in_command(
    command: str,
    merged_config: dict[KeyType, Any],
) -> None:
    """Validate that all params keys in merged config are used in the command.

    Args:
        command: The command template from framework.yml
        merged_config: The final merged configuration

    Raises:
        MisconfigurationError: If merged_config contains params keys not used in command
    """
    # Extract params keys used in command
    command_standard_params, command_extra_params = extract_params_from_command(command)

    # Get params from merged config
    config_params = merged_config.get("config", {}).get("params", {})

    if not config_params:
        return  # No params to validate

    # Check standard params
    unused_standard = []
    for key, value in config_params.items():
        if key == "extra":
            continue  # Handle extra separately
        # Only validate non-None values (None means not set/using default)
        if value is not None and key not in command_standard_params:
            unused_standard.append(f"config.params.{key}")

    # Check params.extra
    config_extra = config_params.get("extra", {})
    unused_extra = []
    for key in config_extra.keys():
        if key not in command_extra_params:
            unused_extra.append(f"config.params.extra.{key}")

    # Raise error if any unused params found
    all_unused = unused_standard + unused_extra
    if all_unused:
        valid_standard = [f"config.params.{p}" for p in sorted(command_standard_params)]
        valid_extra = [f"config.params.extra.{p}" for p in sorted(command_extra_params)]
        logger.warn(
            f"Configuration contains parameter(s) that are not used in the command template: "
            f"{', '.join(all_unused)}. "
            f"Valid params from command: {valid_standard + valid_extra}. "
            f"Remove the unused parameters or update the command template to use them."
        )


def dotlist_to_dict(dotlist: list[str]) -> dict:
    """Resolve dot-list style key-value pairs with YAML.

    Helper for overriding configuration values using command-line arguments in dot-list style.
    """
    dotlist_dict = {}
    for override in dotlist:
        parts = override.strip().split("=", 1)
        if len(parts) == 2:
            key = parts[0].strip()
            raw_value = parts[1].strip()

            # If the value starts with a quote but doesn't end with the same quote,
            # it means we have a malformed string. In this case, we'll treat it as a raw string.
            if (raw_value.startswith('"') and not raw_value.endswith('"')) or (
                raw_value.startswith("'") and not raw_value.endswith("'")
            ):
                value = raw_value
            else:
                try:
                    value = yaml.safe_load(raw_value)
                except yaml.YAMLError:
                    # If YAML parsing fails, treat it as a raw string
                    value = raw_value

            keys = key.split(".")
            temp = dotlist_dict
            for k in keys[:-1]:
                temp = temp.setdefault(k, {})
            temp[keys[-1]] = value
    return dotlist_dict


def run_command(command, cwd=None, verbose=False, propagate_errors=False):
    if verbose:
        logger.info(f"Running command: {command}")
        if cwd:
            print(f"Current working directory set to: {cwd}")

    with tempfile.TemporaryDirectory() as tmpdirname:
        if verbose:
            logger.info(f"Temporary directory created at: {tmpdirname}")

        file = os.path.join(
            tmpdirname, hashlib.sha1(command.encode("utf-8")).hexdigest() + ".sh"
        )
        if verbose:
            logger.info(f"Script file created: {file}")

        with open(file, "w") as f:
            f.write(command)
            f.flush()
            if verbose:
                logger.info("Command written to script file.")

        master, slave = os.openpty()
        process = subprocess.Popen(
            f"bash {file}",
            stdout=slave,
            stderr=slave,
            stdin=subprocess.PIPE,
            cwd=cwd,
            shell=True,
            executable="/bin/bash",
        )

        if verbose:
            logger.info("Subprocess started.")

        os.close(slave)

        if propagate_errors:
            stderr_output = []

        while True:
            try:
                output = os.read(master, 1024)
                if not output:
                    break
                decoded_output = output.decode(errors="ignore")
                print(decoded_output, end="", flush=True)

                if propagate_errors:
                    stderr_output.append(decoded_output)

            except OSError as e:
                if e.errno == 5:  # Input/output error is expected at the end of output
                    break
                raise

        if verbose:
            logger.info("Output reading completed.")

        rc = process.wait()

        if verbose:
            logger.info(f"Subprocess finished with return code: {rc}")

        # New error propagation logic
        if rc != 0 and propagate_errors:
            error_content = (
                "".join(stderr_output) if stderr_output else "No error details captured"
            )
            raise RuntimeError(
                f"Evaluation failed! Please consult the logs below:\n{error_content}"
            )

        return rc


def check_health(
    health_url: str, max_retries: int = 600, retry_interval: int = 2
) -> bool:
    """
    Check the health of the server.
    """
    for _ in range(max_retries):
        try:
            response = requests.get(health_url)
            if response.status_code == 200:
                return True
            logger.info(f"Server replied with status code: {response.status_code}")
            time.sleep(retry_interval)
        except requests.exceptions.RequestException:
            logger.info("Server is not ready")
            time.sleep(retry_interval)
    return False


def check_endpoint(
    endpoint_url: str,
    endpoint_type: Literal["completions", "chat"],
    model_name: str,
    max_retries: int = 600,
    retry_interval: int = 2,
) -> bool:
    """Checks if the OpenAI-compatible endpoint is alive by sending a simple prompt.

    Args:
        endpoint_url (str): Full endpoint URL. For most servers that means either ``/v1/chat/completions`` or ``/completions`` must be provided
        endpoint_type (Literal[completions, chat]): indicates if the model is instruction-tuned (chat) or a base model (completions). Used to constuct a proper payload structure.
        model_name (str): model name that is linked to payload. Might be required by some endpoint.
        max_retries (int, optional): How many attempt before returning false. Defaults to 600.
        retry_interval (int, optional): How many seconds to wait between attempts. Defaults to 2.

    Raises:
        ValueError: if endpoint_type was not one of "completions", "chat"

    Returns:
        bool: whether the endpoint is alive
    """
    payload = {"model": model_name, "max_tokens": 1}
    if endpoint_type == "completions":
        payload["prompt"] = "hello, my name is"
    elif endpoint_type == "chat":
        payload["messages"] = [{"role": "user", "content": "hello, what is your name?"}]
    else:
        raise ValueError(f"Invalid endpoint type: {endpoint_type}")

    for _ in range(max_retries):
        try:
            response = requests.post(endpoint_url, json=payload)
            if response.status_code == 200:
                return True
            logger.info(f"Server replied with status code: {response.status_code}")
            time.sleep(retry_interval)
        except requests.exceptions.RequestException:
            logger.info("Server is not ready")
            time.sleep(retry_interval)
    return False



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nemo_evaluator.api import evaluate, show_available_tasks
from nemo_evaluator.api.api_dataclasses import (
    ApiEndpoint,
    ConfigParams,
    EndpointType,
    Evaluation,
    EvaluationConfig,
    EvaluationResult,
    EvaluationTarget,
    GroupResult,
    MetricResult,
    Score,
    ScoreStats,
    TaskResult,
)
from nemo_evaluator.package_info import (
    __contact_emails__,
    __contact_names__,
    __download_url__,
    __homepage__,
    __package_name__,
    __repository_url__,
    __version__,
)

__all__ = [
    "ApiEndpoint",
    "ConfigParams",
    "EndpointType",
    "Evaluation",
    "EvaluationConfig",
    "EvaluationResult",
    "EvaluationTarget",
    "GroupResult",
    "MetricResult",
    "Score",
    "ScoreStats",
    "TaskResult",
    "run_eval",
    "evaluate",
    "show_available_tasks",
    "__version__",
    "__package_name__",
    "__contact_names__",
    "__contact_emails__",
    "__homepage__",
    "__repository_url__",
    "__download_url__",
]

# Import logging to ensure centralized logging is configured
from nemo_evaluator import logging  # noqa: F401



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Endpoint interceptor that makes actual requests to the upstream API."""

import json
import time
from typing import final

import requests

from nemo_evaluator.adapters.decorators import register_for_adapter
from nemo_evaluator.adapters.types import (
    AdapterGlobalContext,
    AdapterRequest,
    AdapterResponse,
    RequestToResponseInterceptor,
)
from nemo_evaluator.logging import BaseLoggingParams, get_logger


@register_for_adapter(
    name="endpoint",
    description="Makes the actual request to the upstream API",
)
@final
class EndpointInterceptor(RequestToResponseInterceptor):
    """Required interceptor that handles the actual API communication. This interceptor must be present in every configuration as it performs the final request to the target API endpoint.
    Important: This interceptor should always be placed after the last request interceptor and before the first response interceptor."""

    class Params(BaseLoggingParams):
        """Configuration parameters for endpoint interceptor."""

        pass

    def __init__(self, params: Params):
        """
        Initialize the endpoint interceptor.

        Args:
            params: Configuration parameters
        """
        # Get logger for this interceptor with interceptor context
        self.logger = get_logger(self.__class__.__name__)

        self.logger.info("Endpoint interceptor initialized")

    def intercept_request(
        self, ar: AdapterRequest, context: AdapterGlobalContext
    ) -> AdapterResponse:
        """Make the actual request to the upstream API.

        Args:
            ar: The adapter request
            context: Global context containing server-level configuration

        Returns:
            AdapterResponse with the response from the upstream API
        """
        self.logger.debug(
            "Making request to upstream API",
            method=ar.r.method,
            url=context.url,
            headers_count=len(ar.r.headers),
            has_json=ar.r.json is not None,
        )

        # Record start time for latency tracking
        start_time = time.time()

        # This is a final interceptor, we'll need the flask_request and api
        raw_response = requests.request(
            method=ar.r.method,
            url=context.url,
            headers={k: v for k, v in ar.r.headers if k.lower() != "host"},
            json=ar.r.json,
            cookies=ar.r.cookies,
            allow_redirects=False,
        )

        # replace choices[xx].message.content=None with empty string
        if raw_response.content is not None:
            try:
                response_json = json.loads(raw_response.content)
                if (
                    "choices" in response_json
                    and isinstance(response_json["choices"], list)
                    and len(response_json["choices"]) > 0
                ):
                    for i, choice in enumerate(response_json["choices"]):
                        if (
                            "message" in choice
                            and "content" in choice["message"]
                            and choice["message"]["content"] is None
                        ):
                            self.logger.warning(
                                f"choices[{i}].message.content is None, replacing with empty string"
                            )
                            choice["message"]["content"] = ""
                raw_response._content = json.dumps(response_json).encode("utf-8")
            except (json.JSONDecodeError, TypeError, KeyError) as e:
                # If JSON parsing fails or unexpected structure, leave response unchanged
                self.logger.debug(
                    "Could not parse response as JSON, leaving unchanged", error=str(e)
                )

        resp = AdapterResponse(
            r=raw_response,
            rctx=ar.rctx,
            latency_ms=round(
                (time.time() - start_time) * 1000, 2
            ),  # Convert to milliseconds with 2 decimal places
        )

        self.logger.debug(
            "Upstream API request completed",
            status_code=resp.r.status_code,
            reason=resp.r.reason,
            response_headers_count=len(resp.r.headers),
            response_content_length=len(resp.r.content) if resp.r.content else 0,
            latency_ms=resp.latency_ms,
        )

        return resp



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Progress tracking interceptor that tracks number of samples processed via webhook."""

import os
import pathlib
import threading
import time
from typing import Annotated, Optional, final

import requests
from pydantic import Field

from nemo_evaluator.adapters.decorators import register_for_adapter
from nemo_evaluator.adapters.types import (
    AdapterGlobalContext,
    AdapterResponse,
    PostEvalHook,
    ResponseInterceptor,
)
from nemo_evaluator.logging import BaseLoggingParams, get_logger


@register_for_adapter(
    name="progress_tracking",
    description="Tracks number of samples processed via webhook",
)
@final
class ProgressTrackingInterceptor(ResponseInterceptor, PostEvalHook):
    """Progress tracking via external webhook."""

    class Params(BaseLoggingParams):
        """Configuration parameters for progress tracking interceptor."""

        progress_tracking_url: Optional[str] = Field(
            default="http://localhost:8000",
            description="URL to post the number of processed samples to. Supports expansion of shell variables if present.",
        )
        progress_tracking_interval: Annotated[int, Field(gt=0)] = Field(
            default=1,
            description="How often (every how many samples) to send a progress information.",
        )
        progress_tracking_interval_seconds: Optional[
            Annotated[float | None, Field(gt=0)]
        ] = Field(
            default=None,
            description="How often (every N seconds) to send a progress information in addition to progress_tracking_interval.",
        )
        request_method: str = Field(
            default="PATCH",
            description="Request method to use for updating the evaluation progress.",
        )
        output_dir: Optional[str] = Field(
            default=None,
            description="Evaluation output directory. If provided, the progress tracking will be saved to a file in this directory.",
        )

    progress_tracking_url: Optional[str]
    progress_tracking_interval: int
    progress_filepath: Optional[pathlib.Path]
    request_method: str

    def __init__(self, params: Params):
        """
        Initialize the progress tracking interceptor.

        Args:
            params: Configuration parameters
        """
        self.progress_tracking_url = os.path.expandvars(params.progress_tracking_url)
        self.progress_tracking_interval = params.progress_tracking_interval
        self.request_method = params.request_method
        if params.output_dir is not None:
            output_dir = pathlib.Path(params.output_dir)
            output_dir.mkdir(parents=True, exist_ok=True)
            self.progress_filepath = output_dir / "progress"
        else:
            self.progress_filepath = None
        self._samples_processed = self._initialize_samples_processed()
        self._last_updated_samples_processed = self._samples_processed
        self._lock = threading.Lock()

        # Get logger for this interceptor with interceptor context
        self.logger = get_logger(self.__class__.__name__)

        # Optional update on timer
        self.progress_tracking_interval_seconds = (
            params.progress_tracking_interval_seconds
        )
        if self.progress_tracking_interval_seconds:
            self._timer_stopped = False
            self._update_on_timer_thread = threading.Thread(
                target=self._update_on_timer,
                kwargs={"interval_seconds": self.progress_tracking_interval_seconds},
                daemon=True,
            )
            self._update_on_timer_thread.start()

        self.logger.info(
            "Progress tracking interceptor initialized",
            progress_tracking_url=self.progress_tracking_url,
            progress_tracking_interval=self.progress_tracking_interval,
            progress_tracking_interval_seconds=self.progress_tracking_interval_seconds,
            output_dir=str(self.progress_filepath) if self.progress_filepath else None,
            initial_samples_processed=self._samples_processed,
        )

    def _initialize_samples_processed(self) -> int:
        if self.progress_filepath is not None and self.progress_filepath.exists():
            with open(self.progress_filepath, "r") as f:
                try:
                    return int(f.read())
                except ValueError:
                    self.logger.warning(
                        "Failed to read progress from file, starting from 0",
                        filepath=str(self.progress_filepath),
                    )
                    return 0
        return 0

    def _write_progress(self, num_samples: int):
        with self._lock:
            self.progress_filepath.write_text(str(num_samples))
            self.logger.debug(
                "Progress written to file",
                filepath=str(self.progress_filepath),
                samples_processed=num_samples,
            )

    def _send_progress(self, num_samples: int) -> requests.Response:
        self.logger.debug(
            "Sending progress to tracking server",
            url=self.progress_tracking_url,
            method=self.request_method,
            samples_processed=num_samples,
        )
        body = {"samples_processed": num_samples}
        try:
            resp = requests.request(
                self.request_method,
                self.progress_tracking_url,
                json=body,
            )
            if resp.status_code >= 200 and resp.status_code < 300:
                self.logger.debug(
                    "Progress sent successfully", samples_processed=num_samples
                )
            else:
                self.logger.warning(
                    "Failed to update job progress",
                    body=body,
                    status_code=resp.status_code,
                    response_text=resp.text,
                )
            return resp
        except requests.exceptions.RequestException as e:
            self.logger.error(
                "Failed to communicate with progress tracking server",
                error=str(e),
                samples_processed=num_samples,
            )

    def _update_on_timer(self, interval_seconds: float):
        """
        Sends an update on a timed interval if there has been a change since the last update.
        This is a blocking function that is expected to be executed in a thread.
        """
        assert interval_seconds > 0
        while True:
            time.sleep(interval_seconds)
            with self._lock:
                if self._timer_stopped:
                    return
                if self._last_updated_samples_processed == self._samples_processed:
                    continue
                curr_samples = self._samples_processed

            if self.progress_tracking_url is not None:
                self._send_progress(curr_samples)
            if self.progress_filepath is not None:
                self._write_progress(curr_samples)

            self.logger.info(
                "Progress milestone updated on time interval",
                samples_processed=curr_samples,
                interval=self.progress_tracking_interval,
            )
            with self._lock:
                self._last_updated_samples_processed = curr_samples

    @final
    def intercept_response(
        self, ar: AdapterResponse, context: AdapterGlobalContext
    ) -> AdapterResponse:
        curr_samples = 0
        with self._lock:
            self._samples_processed += 1
            curr_samples = self._samples_processed

        self.logger.debug(
            "Sample processed",
            samples_processed=curr_samples,
            interval=self.progress_tracking_interval,
        )

        if (curr_samples % self.progress_tracking_interval) == 0:
            if self.progress_tracking_url is not None:
                self._send_progress(curr_samples)
            if self.progress_filepath is not None:
                self._write_progress(curr_samples)

            self.logger.info(
                "Progress milestone reached",
                samples_processed=curr_samples,
                interval=self.progress_tracking_interval,
            )
            with self._lock:
                self._last_updated_samples_processed = curr_samples

        return ar

    def post_eval_hook(self, context: AdapterGlobalContext) -> None:
        self.logger.info(
            "Post-eval hook executed", total_samples_processed=self._samples_processed
        )
        with self._lock:
            if self.progress_tracking_interval_seconds:
                self._timer_stopped = True
            if self._samples_processed == self._last_updated_samples_processed:
                return

        if self.progress_tracking_url is not None:
            self._send_progress(self._samples_processed)
        if self.progress_filepath is not None:
            self._write_progress(self._samples_processed)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import pkgutil
import sys

import yaml

# Import logging to ensure centralized logging is configured
from nemo_evaluator import logging  # noqa: F401
from nemo_evaluator.api.api_dataclasses import (
    EvaluationConfig,
    EvaluationMetadata,
    EvaluationTarget,
)
from nemo_evaluator.core.evaluate import evaluate
from nemo_evaluator.core.input import (
    _get_framework_evaluations,
    _parse_cli_args,
    load_run_config,
    validate_configuration,
)

__all__ = []

from .utils import deep_update


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--debug", action="store_true", help="Debug the core_evals script"
    )
    subparsers = parser.add_subparsers(help="Functions")
    parser_ls = subparsers.add_parser("ls", help="List available evaluation types")
    parser_ls.set_defaults(command="ls")

    parser_run = subparsers.add_parser("run_eval", help="Run the evaluation")
    parser_run.add_argument("--eval_type", type=str, help="Run config.: task name")
    parser_run.add_argument("--model_id", type=str, help="Run config.: model name")
    parser_run.add_argument(
        "--model_type",
        type=str,
        help="Run config.: endpoint type",
        choices=["chat", "completions", "vlm", "embedding"],
    )
    parser_run.add_argument("--model_url", type=str, help="Run config.: model URI")
    parser_run.add_argument(
        "--output_dir", type=str, help="Run config.: results output dir."
    )
    parser_run.add_argument(
        "--api_key_name",
        type=str,
        help="Run config.: API key env variable name (optional)",
        default=None,
    )
    parser_run.add_argument(
        "--run_config",
        type=str,
        help="Load the run configuration from the YAML file (optional and overridden by the cli arguments)",
        default=None,
    )
    parser_run.add_argument(
        "--overrides",
        type=str,
        help="Comma-separated dot-style parameters to override config values (overriding values from run_config and CLI args)",
        default=None,
    )
    parser_run.add_argument(
        "--dry_run",
        action="store_true",
        help="Shows rendered config and command instead of running",
        default=False,
    )
    parser_run.set_defaults(command="run_eval")
    return parser


def get_args() -> argparse.Namespace:
    parser = get_parser()

    args = parser.parse_args()

    if args.debug:
        # Override with debug level if --debug flag is used
        from nemo_evaluator.logging import get_logger

        logger = get_logger(__name__)
        logger.warning(
            "This flag is deprecated and will be removed in the future, please set environment variable NEMO_EVALUATOR_LOG_LEVEL=DEBUG instead!"
        )
        logger.warning("Setting NEMO_EVALUATOR_LOG_LEVEL=DEBUG")
        os.environ["NEMO_EVALUATOR_LOG_LEVEL"] = "DEBUG"

    if "command" not in args:
        parser.print_help()
        sys.exit(0)
    return args


def show_available_tasks() -> None:
    """
    Prints all available evaluations in the format of::

        {harness1}:
            * benchmark A
            * benchmark B
        {harness2}:
            * benchmark A
            * benchmark B
        ...

    .. important:: Only evaluations from installed wheels are being displayed.
    """
    try:
        import core_evals

        core_evals_pkg = list(pkgutil.iter_modules(core_evals.__path__))
    except ImportError:
        core_evals_pkg = []

    if not core_evals_pkg:
        print("NO evaluation packages are installed.")

    for pkg in core_evals_pkg:
        framework_eval_mapping, *_ = _get_framework_evaluations(
            os.path.join(pkg.module_finder.path, pkg.name, "framework.yml")
        )
        for ind_pkg in framework_eval_mapping.keys():
            print(f"{ind_pkg}: ")
            for task in framework_eval_mapping[ind_pkg].keys():
                print(f"  * {task}")


def run(args) -> None:
    run_config = load_run_config(args.run_config) if args.run_config else {}
    # CLI args take precedence over YAML run config
    run_config = deep_update(run_config, _parse_cli_args(args), skip_nones=True)
    if args.dry_run:
        evaluation = validate_configuration(run_config)
        print("Rendered config:\n")
        config = evaluation.model_dump()
        print(yaml.dump(config, sort_keys=False, default_flow_style=False, indent=2))
        print("\nRendered command:\n")
        cmd = evaluation.render_command()
        print(cmd)
        exit(0)

    # Validate configuration first (catches config errors early, before evaluate())
    validate_configuration(run_config)

    metadata_cfg: EvaluationMetadata | None = run_config.get("metadata")

    # Build evaluation configuration with framework defaults merged
    evaluation = validate_configuration(run_config)

    eval_cfg = EvaluationConfig(**evaluation.model_dump()["config"])
    target_cfg = EvaluationTarget(**evaluation.model_dump()["target"])

    evaluate(eval_cfg=eval_cfg, target_cfg=target_cfg, metadata=metadata_cfg)


def run_eval() -> None:
    """
    CLI entry point for running evaluations.

    This function parses command line arguments and executes evaluations.
    It does not take parameters directly - all configuration is passed via CLI arguments.

    CLI Arguments:
        --eval_type: Type of evaluation to run (e.g., "mmlu_pro", "gsm8k")
        --model_id: Model identifier (e.g "meta/llama-3.1-8b-instruct")
        --model_url: API endpoint URL (e.g "https://integrate.api.nvidia.com/v1/chat/completions" for chat endpoint type)
        --model_type: Endpoint type ("chat", "completions", "vlm", "embedding")
        --api_key_name: Environment variable name for API key integration with endpoints (optional)
        --output_dir: Output directory for results
        --run_config: Path to YAML Run Configuration file (optional)
        --overrides: Comma-separated dot-style parameter overrides (optional)
        --dry_run: Show rendered config without running (optional)
        --debug: Enable debug logging (optional, deprecated, use NV_LOG_LEVEL=DEBUG env var)

    Usage:
        run_eval()  # Parses sys.argv automatically
    """
    args = get_args()

    if sys.argv[0].endswith("eval-factory"):
        from nemo_evaluator.logging import get_logger

        logger = get_logger(__name__)
        logger.warning(
            "You appear to be using a deprecated eval_factory command. Please use nemo-evaluator instead with the same arguments. eval-factory command is going to be removed before 25.12 containers are released."
        )

    if args.command == "ls":
        show_available_tasks()
    elif args.command == "run_eval":
        run(args)


if __name__ == "__main__":
    run_eval()
